package supply_analysis

import (
	"errors"
	"eta_gn/eta_index_lib/global"
	"eta_gn/eta_index_lib/utils"
	"github.com/shopspring/decimal"
	"gorm.io/gorm"
	"time"
)

type BaseFromStockPlantData struct {
	StockPlantDataId int       `gorm:"primaryKey;autoIncrement;column:stock_plant_data_id" description:"指标数据ID"`
	VarietyEdbId     int       `gorm:"column:variety_edb_id" description:"指标id"`
	DataTime         time.Time `gorm:"column:data_time" description:"数据日期"`
	Value            float64   `gorm:"column:value" description:"数据值"`
	ModifyTime       time.Time `gorm:"column:modify_time" description:"修改时间"`
	CreateTime       time.Time `gorm:"column:create_time" description:"创建时间"`
	DataTimestamp    int64     `gorm:"column:data_timestamp" description:"数据日期时间戳"`
}

type CalculateVarietyReq struct {
	VarietyId int    `description:"品种id"`
	AdminId   int    `description:"操作人id"`
	AdminName string `description:"操作人名称"`
}

func Calculate(varietyId int, sysUserId int, sysUserName string) (err error) {
	varietyInfo, err := GetVarietyById(varietyId)
	if err != nil {
		return
	}

	plantList, err := GetAllVarietyPlantByVarietyId(varietyId)
	if err != nil {
		return
	}

	varietyEdbInfoList, err := GetAllVarietyEdbInfoByVarietyId(varietyId)
	if err != nil {
		return
	}

	var lastDate time.Time
	{
		currDate := time.Now()
		currHour := currDate.Hour()
		if currHour < 22 {
			lastDate, _ = time.ParseInLocation(utils.FormatDate, currDate.Format(utils.FormatDate), time.Local)
		} else {
			lastDate, _ = time.ParseInLocation(utils.FormatDate, currDate.AddDate(0, 0, 1).Format(utils.FormatDate), time.Local)
		}
		lastDate = lastDate.AddDate(0, 0, 365) //业务需要计算的数据日期往当前日期后面加上一年
	}

	var startDate time.Time
	maintenanceDataMap := make(map[time.Time][]float64)

	for _, plantInfo := range plantList {
		if plantInfo.IsStop == 1 { // 如果是停产了，那么结束日期就是到 产生的数据的结束日期
			plantInfo.ResumptionDate = lastDate
		}
		if plantInfo.MaintenanceDate.IsZero() || plantInfo.ResumptionDate.IsZero() {
			continue
		}
		if plantInfo.AverageDailyCapacityReduction == 0 {
			continue
		}
		if startDate.IsZero() || plantInfo.MaintenanceDate.Before(startDate) {
			startDate = plantInfo.MaintenanceDate
		}

		resumptionDate := plantInfo.ResumptionDate

		for tmpDate := plantInfo.MaintenanceDate; !tmpDate.After(resumptionDate); tmpDate = tmpDate.AddDate(0, 0, 1) {
			dataList, ok := maintenanceDataMap[tmpDate]
			if !ok {
				dataList = make([]float64, 0)
			}
			dataList = append(dataList, plantInfo.AverageDailyCapacityReduction)
			maintenanceDataMap[tmpDate] = dataList
		}
	}

	if len(maintenanceDataMap) <= 0 {
		return
	}

	dataMap := make(map[time.Time]float64)
	for tmpDate := startDate; !tmpDate.After(lastDate); tmpDate = tmpDate.AddDate(0, 0, 1) {
		var calculateVal float64
		tmpDataList, ok := maintenanceDataMap[tmpDate]
		if !ok {
			calculateVal = 0
		} else {
			for _, tmpVal := range tmpDataList {
				calculateVal += tmpVal
			}
		}

		dataMap[tmpDate] = calculateVal
	}

	to := global.DEFAULT_DmSQL.Begin()
	defer func() {
		if err != nil {
			to.Rollback()
		} else {
			to.Commit()
		}
	}()

	if len(dataMap) > 0 {
		for _, varietyEdbInfo := range varietyEdbInfoList {
			err = calculateEdb(to, varietyEdbInfo, dataMap, startDate, lastDate)
			if err != nil {
				return
			}
		}
	}

	varietyInfo.LastUpdateSysUserId = sysUserId
	varietyInfo.LastUpdateSysUserRealName = sysUserName
	varietyInfo.ModifyTime = time.Now()
	err = to.Model(varietyInfo).Select([]string{"LastUpdateSysUserId", "LastUpdateSysUserRealName", "ModifyTime"}).Updates(varietyInfo).Error

	return
}

func calculateEdb(to *gorm.DB, varietyEdbInfo *VarietyEdbInfo, dataMap map[time.Time]float64, startDate, lastDate time.Time) (err error) {
	sql := `SELECT * FROM base_from_stock_plant_data WHERE variety_edb_id = ? `
	var list []*BaseFromStockPlantData
	err = to.Raw(sql, varietyEdbInfo.VarietyEdbId).Scan(&list).Error
	if err != nil {
		return
	}

	addList := make([]*BaseFromStockPlantData, 0)
	deleteIdMap := make(map[int]int, 0)

	existData := make(map[time.Time]*BaseFromStockPlantData)
	for _, v := range list {
		existData[v.DataTime] = v
		deleteIdMap[v.StockPlantDataId] = v.StockPlantDataId
	}

	endDate := lastDate     // 实际数据最后结束的日期
	var latestValue float64 // 最新值

	switch varietyEdbInfo.Source {
	case 1, 2:
		weekDataMap := make(map[time.Time]float64)

		for currDate := startDate; !currDate.After(lastDate); currDate = currDate.AddDate(0, 0, 1) {
			var weekVal, currVal float64
			for i := 0; i <= 6; i++ {
				tmpDate := currDate.AddDate(0, 0, -i)
				tmpData, ok := dataMap[tmpDate]
				if !ok {
					continue
				}
				weekVal = weekVal + tmpData
			}
			weekDataMap[currDate] = weekVal // 周度产量数据

			if varietyEdbInfo.Source == 1 {
				currVal = weekVal
			} else {
				preVal, ok := weekDataMap[currDate.AddDate(0, 0, -7)]
				if !ok {
					preVal = 0
				}
				currVal = preVal - weekVal
			}

			currVal, _ = decimal.NewFromFloat(currVal).Round(4).Float64()
			latestValue = currVal
			timestamp := currDate.UnixNano() / 1e6
			existItem, ok := existData[currDate]
			if !ok {
				addList = append(addList, &BaseFromStockPlantData{
					VarietyEdbId:  varietyEdbInfo.VarietyEdbId,
					DataTime:      currDate,
					Value:         currVal,
					ModifyTime:    time.Now(),
					CreateTime:    time.Now(),
					DataTimestamp: timestamp,
				})
			} else {
				if existItem.Value != currVal {
					existItem.Value = currVal
					existItem.ModifyTime = time.Now()
					err = to.Model(existItem).Select([]string{"Value", "ModifyTime"}).Updates(existItem).Error
					if err != nil {
						return
					}
				}

				delete(deleteIdMap, existItem.StockPlantDataId)
			}
		}
	case 3, 4:
		tmpDataList := make([]float64, 0)        // 每天的值
		preSumValDeci := decimal.NewFromFloat(0) // 上一期的值

		for tmpDate := startDate; !tmpDate.After(lastDate); tmpDate = tmpDate.AddDate(0, 0, 1) {
			tmpData, ok := dataMap[tmpDate]
			if !ok {
				continue
			}
			tmpDataList = append(tmpDataList, tmpData)

			if tmpDate.AddDate(0, 0, 1).Day() != 1 {
				continue
			}

			timestamp := tmpDate.UnixNano() / 1e6

			var sumVal float64
			for _, tmpVal := range tmpDataList {
				sumVal += tmpVal
			}
			sumValDeci := decimal.NewFromFloat(sumVal)

			var currVal float64
			if varietyEdbInfo.Source == 3 { // 月度影响产量
				currVal, _ = sumValDeci.Round(4).Float64()
			} else { // 如果是月度产量变动的话
				currVal, _ = (preSumValDeci.Sub(decimal.NewFromFloat(sumVal))).Round(4).Float64()
			}
			latestValue = currVal

			preSumValDeci = sumValDeci       // 赋值给上一个周期的截止值
			tmpDataList = make([]float64, 0) //重置值列表，便于下一个周期计算

			existItem, ok := existData[tmpDate]
			if !ok {
				addList = append(addList, &BaseFromStockPlantData{
					VarietyEdbId:  varietyEdbInfo.VarietyEdbId,
					DataTime:      tmpDate,
					Value:         currVal,
					ModifyTime:    time.Now(),
					CreateTime:    time.Now(),
					DataTimestamp: timestamp,
				})
			} else {
				if existItem.Value != currVal {
					existItem.Value = currVal
					existItem.ModifyTime = time.Now()
					err = to.Model(existItem).Select([]string{"Value", "ModifyTime"}).Updates(existItem).Error
					if err != nil {
						return
					}
				}

				delete(deleteIdMap, existItem.StockPlantDataId)
			}
		}

		currMonthEndDay := time.Date(lastDate.Year(), lastDate.Month(), 1, 0, 0, 0, 0, time.Local).AddDate(0, 1, -1)
		if !lastDate.Equal(currMonthEndDay) {
			tmpDate := currMonthEndDay
			endDate = tmpDate //数据实际结束的日期

			existItem, ok := existData[tmpDate]

			timestamp := tmpDate.UnixNano() / 1e6
			var sumVal float64
			for _, tmpVal := range tmpDataList {
				sumVal += tmpVal
			}
			sumValDeci := decimal.NewFromFloat(sumVal)

			var currVal float64
			if varietyEdbInfo.Source == 3 { // 月度影响产量
				currVal, _ = sumValDeci.Round(4).Float64()
			} else { // 如果是月度产量变动的话
				currVal, _ = (preSumValDeci.Sub(decimal.NewFromFloat(sumVal))).Round(4).Float64()
			}
			currVal, _ = decimal.NewFromFloat(currVal).Round(4).Float64()
			latestValue = currVal

			if !ok {
				addList = append(addList, &BaseFromStockPlantData{
					VarietyEdbId:  varietyEdbInfo.VarietyEdbId,
					DataTime:      tmpDate,
					Value:         currVal,
					ModifyTime:    time.Now(),
					CreateTime:    time.Now(),
					DataTimestamp: timestamp,
				})
			} else {
				if existItem.Value != currVal {
					existItem.Value = currVal
					existItem.ModifyTime = time.Now()
					err = to.Model(existItem).Select([]string{"Value", "ModifyTime"}).Updates(existItem).Error
					if err != nil {
						return
					}
				}

				delete(deleteIdMap, existItem.StockPlantDataId)
			}
		}

	default:
		err = errors.New("错误的指标类型")
		return
	}

	lenAddList := len(addList)
	if lenAddList > 0 {
		err = to.CreateInBatches(addList, 500).Error
		if err != nil {
			return
		}
	}

	lenDeleteIdMap := len(deleteIdMap)
	if len(deleteIdMap) > 0 {
		deleteIdList := make([]int, 0)
		for _, v := range deleteIdMap {
			deleteIdList = append(deleteIdList, v)
		}
		sql = ` DELETE FROM base_from_stock_plant_data WHERE stock_plant_data_id in (` + utils.GetOrmInReplace(lenDeleteIdMap) + `) `
		err = to.Exec(sql, deleteIdList).Error
		if err != nil {
			return
		}
	}

	varietyEdbInfo.StartDate = startDate
	varietyEdbInfo.EndDate = endDate
	varietyEdbInfo.LatestValue = latestValue
	varietyEdbInfo.ModifyTime = time.Now()
	err = to.Model(varietyEdbInfo).Select([]string{"StartDate", "EndDate", "LatestValue", "ModifyTime"}).Updates(varietyEdbInfo).Error
	return
}
